from django.dispatch import Signal, receiver
from django.utils import timezone

from django_ses.signals import bounce_received, complaint_received, delivery_received

from .models import SESMailDelivery


# our signals
recipients_bounce = Signal(providing_args=['recipients', 'bounce_obj'])
recipients_complaint = Signal(providing_args=['recipients', 'complaint_obj'])
recipients_delivery = Signal(providing_args=['recipients', 'delivery_obj'])


# handle django-ses signals

@receiver(bounce_received)
def bounce_handler(sender, mail_obj, bounce_obj, raw_message, **kwargs):
    recipients = [r['emailAddress'] for r in bounce_obj['bouncedRecipients']]
    SESMailDelivery.objects.filter(message_id=mail_obj['messageId'], recipient__in=recipients).update(
        state=SESMailDelivery.STATE_BOUNCED,
        state_data=bounce_obj,
        mail_data=mail_obj,
        updated_at=timezone.now(),
    )
    recipients_bounce.send(__name__, recipients=recipients, bounce_obj=bounce_obj)


@receiver(complaint_received)
def complaint_handler(sender, mail_obj, complaint_obj, raw_message, **kwargs):
    recipients = [r['emailAddress'] for r in complaint_obj['complainedRecipients']]
    SESMailDelivery.objects.filter(message_id=mail_obj['messageId'], recipient__in=recipients).update(
        state=SESMailDelivery.STATE_COMPLAINT,
        state_data=complaint_obj,
        mail_data=mail_obj,
        updated_at=timezone.now(),
    )
    recipients_complaint.send(__name__, recipients=recipients, complaint_obj=complaint_obj)


@receiver(delivery_received)
def delivery_handler(sender, mail_obj, delivery_obj, raw_message, **kwargs):
    recipients = delivery_obj['recipients']
    SESMailDelivery.objects.filter(message_id=mail_obj['messageId'], recipient__in=recipients).update(
        state=SESMailDelivery.STATE_DELIVERED,
        state_data=delivery_obj,
        mail_data=mail_obj,
        updated_at=timezone.now(),
    )
    recipients_delivery.send(__name__, recipients=recipients, delivery_obj=delivery_obj)



from django.conf import settings
from django.utils.module_loading import import_string

from django_ses import SESBackend

from .models import SESMailDelivery


class SESSNSTrackerBackend(SESBackend):
    def send_messages(self, email_messages) -> int:
        if settings.DEBUG and getattr(settings, 'SES_SNS_TRACKER_DEBUG_BACKEND', None):
            debug_backend = import_string(settings.SES_SNS_TRACKER_DEBUG_BACKEND)
            num_sent = debug_backend().send_messages(email_messages)
        else:
            num_sent = super().send_messages(email_messages)
        deliveries = list()

        for message in email_messages:
            for recipient in message.recipients():
                deliveries.append(SESMailDelivery(
                    recipient=recipient,
                    message_id=message.extra_headers.get('message_id', 'NO_MESSAGE_ID'),
                    request_id=message.extra_headers.get('request_id', 'NO_RESULT_ID'),
                ))

        SESMailDelivery.objects.bulk_create(deliveries)

        return num_sent



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ses_sns_tracker', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='sesmaildelivery',
            options={'verbose_name': 'SES Mail Delivery', 'verbose_name_plural': 'SES Mail Deliveries'},
        ),
    ]



from django.conf import settings

from django_ses.views import SESEventWebhookView

from .utils import cryptography_verify_event_message


class SESSNSTrackerWebhookView(SESEventWebhookView):
    def verify_event_message(self, notification):
        if getattr(settings, 'SES_SNS_TRACKER_USE_CRYPTOGRAPHY', True):
            return cryptography_verify_event_message(notification)
        return super().verify_event_message(notification)



from django.contrib import admin

from .models import SESMailDelivery


@admin.register(SESMailDelivery)
class SESMailDeliveryAdmin(admin.ModelAdmin):
    list_display = ('recipient', 'sent_at', 'updated_at', 'state')
    list_filter = ('state',)
    readonly_fields = ('sent_at', 'updated_at')
    ordering = ('-updated_at',)



from django.apps import AppConfig


class SESSNSTrackerConfig(AppConfig):
    name = 'ses_sns_tracker'
    verbose_name = "SES/SNS Tracker"

    def ready(self):
        from . import signals  # NOQA



from django.urls import path

from .views import SESSNSTrackerWebhookView


"""
URLConf left for compatibility. New setups should use the view directly in their own URLConf.
"""

app_name = 'ses_sns_tracker'

urlpatterns = [
    path('bounce/', SESSNSTrackerWebhookView.as_view(), name='handle-event-webhook'),
]



import base64
import logging

import requests
from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding

from django_ses.utils import EventMessageVerifier


logger = logging.getLogger(__name__)


class CryptographyEventMessageVerifier(EventMessageVerifier):
    """Custom EventMessageVerifier using cryptography instead of M2Crypto"""

    def is_verified(self):
        """Verifies an SES bounce message."""
        if self._verified is None:
            signature = self._data.get('Signature')
            if not signature:
                self._verified = False
                return self._verified

            # Decode the signature from base64
            signature = bytes(base64.b64decode(signature))

            # Get the message to sign
            sign_bytes = self._get_bytes_to_sign()
            if not sign_bytes:
                self._verified = False
                return self._verified

            if not self.certificate:
                self._verified = False
                return self._verified

            # Extract the public key
            pkey = self.certificate.public_key()

            # Use the public key to verify the signature.
            try:
                pkey.verify(signature, sign_bytes, padding.PKCS1v15(), hashes.SHA1())
                self._verified = True
            except InvalidSignature:
                logger.debug(u'Failed to validate signature')
                self._verified = False

        return self._verified

    @property
    def certificate(self):
        """Retrieves the certificate used to sign the bounce message."""

        if not hasattr(self, '_certificate'):
            cert_url = self._get_cert_url()

            if not cert_url:
                self._certificate = None
                return self._certificate

            response = requests.get(cert_url)
            if response.status_code != 200:
                logger.warning(u'Could not download certificate from %s: "%s"', cert_url, response.status_code)
                self._certificate = None
                return self._certificate

            try:
                self._certificate = x509.load_pem_x509_certificate(response.content, default_backend())
            except ValueError:
                logger.warning(u'Failed to load certificate from %s', cert_url)
                self._certificate = None

        return self._certificate


def cryptography_verify_event_message(notification):
    """Verify an SES/SNS bounce notification message."""
    verifier = CryptographyEventMessageVerifier(notification)
    return verifier.is_verified()



from typing import Optional

from django.contrib.postgres.fields import JSONField
from django.db import models

from .managers import SESMailManager


class SESMailDelivery(models.Model):
    STATE_SENT = 0
    STATE_DELIVERED = 1
    STATE_BOUNCED = 2
    STATE_COMPLAINT = 3
    DELIVERY_STATES = (
        (STATE_SENT, 'Sent'),
        (STATE_DELIVERED, 'Delivered'),
        (STATE_BOUNCED, 'Bounced'),
        (STATE_COMPLAINT, 'Complaint'),
    )

    recipient = models.EmailField()
    message_id = models.CharField(max_length=128)
    request_id = models.CharField(max_length=128)
    state = models.PositiveSmallIntegerField(choices=DELIVERY_STATES, default=STATE_SENT)
    state_data = JSONField(
        default=dict, blank=True,
        help_text='Amazon SNS event data (bounce/complaint/delivery object)',
    )
    mail_data = JSONField(default=dict, help_text='Amazon SNS mail data', blank=True)
    sent_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    objects = SESMailManager()

    class Meta:
        verbose_name = 'SES Mail Delivery'
        verbose_name_plural = 'SES Mail Deliveries'

    def __str__(self):
        return f'{self.recipient} ({self.get_state_display()} {self.updated_at})'

    @property
    def success(self) -> Optional[bool]:
        if self.state == self.STATE_SENT:
            return None
        return self.state == self.STATE_DELIVERED

    @property
    def error_reason(self) -> Optional[str]:
        if self.state == self.STATE_BOUNCED:
            return self._bounce_reason()
        if self.state == self.STATE_COMPLAINT:
            return self._complaint_reason()

    def _bounce_reason(self) -> str:
        return f'{self.state_data.get("bounceType")}.{self.state_data.get("bounceSubType")}'

    def _complaint_reason(self) -> str:
        return self.state_data.get('complaintFeedbackType')



from typing import TYPE_CHECKING, List

from django.conf import settings
from django.core.mail import EmailMessage
from django.db import models
from django.utils.module_loading import import_string

from django_ses import SESBackend


if TYPE_CHECKING:
    from .models import SESMailDelivery


class SESMailManager(models.Manager):
    def create_message(
        self,
        message: EmailMessage,
        fail_silently: bool = False,
        fake_delivery: bool = False,
    ) -> List['SESMailDelivery']:
        assert isinstance(message, EmailMessage)
        assert message.connection is None or isinstance(message.connection, SESBackend)

        if settings.DEBUG and getattr(settings, 'SES_SNS_TRACKER_DEBUG_BACKEND', None):
            debug_backend = import_string(settings.SES_SNS_TRACKER_DEBUG_BACKEND)
            message.connection = debug_backend()
        elif message.connection is None:
            message.connection = SESBackend()

        if not fake_delivery:
            message.send(fail_silently=fail_silently)

        deliveries = list()
        for recipient in message.recipients():
            deliveries.append(self.model(
                recipient=recipient,
                message_id=message.extra_headers.get('message_id', 'NO_MESSAGE_ID'),
                request_id=message.extra_headers.get('request_id', 'NO_RESULT_ID'),
                state=self.model.STATE_DELIVERED if fake_delivery else self.model.STATE_SENT,
            ))
        return self.bulk_create(deliveries)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-21 11:22
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SESMailDelivery',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('recipient', models.EmailField(max_length=254)),
                ('message_id', models.CharField(max_length=128)),
                ('request_id', models.CharField(max_length=128)),
                ('state', models.PositiveSmallIntegerField(choices=[(0, 'Sent'), (1, 'Delivered'), (2, 'Bounced'), (3, 'Complaint')], default=0)),
                ('state_data', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, help_text='Amazon SNS event data (bounce/complaint/delivery object)')),
                ('mail_data', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, help_text='Amazon SNS mail data')),
                ('sent_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
            options={
                'ordering': ('-updated_at',),
                'verbose_name': 'SES Mail Delivery',
                'verbose_name_plural': 'SES Mail Deliveries',
            },
        ),
    ]
